import os
import re
import sys
import logging

# get parameters from args
parameters = list(filter(
                lambda s: s.find('=') > -1 and bool(re.match(r'[A-Za-z0-9_]*=[.\/A-Za-z0-9]*', s)),
                sys.argv
            ))

# set parameters to env variables
for parameter in parameters:
    variable = parameter.split('=')[0]
    value = parameter.split('=', 1)[-1]
    logging.debug(f'Parameter: {variable} = "{value}"')
    os.environ[variable] = value



import os
import subprocess
import pytest
from typing import Any, Dict, List, Optional, Union
from pathlib import Path
from src.c3.utils import convert_notebook, increase_image_version, get_image_version
from src.c3.pythonscript import Pythonscript

TEST_NOTEBOOK_PATH = 'example_notebook.ipynb'
TEST_SCRIPT_PATH = 'example_script.py'
TEST_RSCRIPT_PATH = 'example_rscript.R'
DUMMY_REPO = 'test'

test_convert_notebook_input = [
    (
        TEST_NOTEBOOK_PATH,
        ['input_path', 'batch_size', 'debug', 'output_path']
    )
]

@pytest.mark.parametrize(
    "notebook_path, env_values",
    test_convert_notebook_input,
)
def test_convert_notebook(
        notebook_path: str,
        env_values: List,
):
    # convert notebook
    script_path = convert_notebook(notebook_path)

    assert os.path.isfile(script_path), f"Error! No file {script_path}"

    # check if script runs with errors
    for env in env_values:
        os.environ[env] = '0'
    subprocess.run(['python', script_path], check=True)

    # check if converted script is processable for create_operator
    py = Pythonscript(script_path)
    name = py.get_name()
    assert isinstance(name, str), "Name is not a string."
    description = py.get_description()
    assert isinstance(description, str), "Description is not a string."
    inputs = py.get_inputs()
    assert isinstance(inputs, dict), "Inputs is not a dict."
    outputs = py.get_outputs()
    assert isinstance(outputs, dict), "Ouputs is not a dict."
    requirements = py.get_requirements()
    assert isinstance(requirements, list), "Requirements is not a list."

    # remove temporary file
    os.remove(script_path)


test_get_remote_version_input = [
    ('us.icr.io/geodn', 'sleep',),
    ('docker.io/romeokienzler', 'predict-image-endpoint',),
]


@pytest.mark.parametrize(
    "repository, name",
    test_get_remote_version_input,
)
def test_get_remote_version(
        repository: str,
        name: str,
):
    # testing icr.io requires 'ibmcloud login'
    version = get_image_version(repository, name)
    assert version != '0.1', \
        f"get_image_version returns default version 0.1"


test_increase_version_input = [
    ('0.1', '0.2'),
    ('2.1.13', '2.1.14'),
    ('0.1beta', '0.1beta.1'),
    ('0.1beta.1', '0.1beta.2'),
]


@pytest.mark.parametrize(
    "last_version, expected_version",
    test_increase_version_input,
)
def test_increase_version(
        last_version: str,
        expected_version: str,
):
    new_version = increase_image_version(last_version)
    assert new_version == expected_version, \
        f"Mismatch between new version {new_version} and expected version {expected_version}"


test_create_operator_input = [
    (
        TEST_SCRIPT_PATH,
        DUMMY_REPO,
        [TEST_NOTEBOOK_PATH],
    ),
    (
        TEST_RSCRIPT_PATH,
        DUMMY_REPO,
        [],
    ),
    (
        TEST_NOTEBOOK_PATH,
        DUMMY_REPO,
        [],
    ),
]
@pytest.mark.parametrize(
    "file_path, repository, args",
    test_create_operator_input,
)
def test_create_operator(
        file_path: str,
        repository: str,
        args: List,
):
    subprocess.run(['python', '../src/c3/create_operator.py', file_path, *args, '-r', repository,
                    '--local_mode', '-v', 'test', '--log_level', 'DEBUG', '--overwrite'],
                   check=True)

    file = Path(file_path)
    file.with_suffix('.yaml').unlink()
    file.with_suffix('.job.yaml').unlink()
    file.with_suffix('.cwl').unlink()
    image_name = f"{repository}/claimed-{file_path.rsplit('.')[0].replace('_', '-')}:test"
    subprocess.run(['docker', 'run', image_name],
                   check=True)


test_create_gridwrapper_input = [
    (
        TEST_SCRIPT_PATH,
        None,
        'process',
        [TEST_NOTEBOOK_PATH],
    ),
    (
        TEST_NOTEBOOK_PATH,
        DUMMY_REPO,
        'your_function',
        [],
    ),
]
@pytest.mark.parametrize(
    "file_path, repository, process, args",
    test_create_gridwrapper_input,
)
def test_create_gridwrapper(
        file_path: str,
        repository: str,
        process: str,
        args: List,
):
    subprocess.run(['python', '../src/c3/create_gridwrapper.py', file_path, *args, '--overwrite',
                    '-p', process, '--local_mode', '-v', 'test', '--log_level', 'DEBUG'], check=True)

    file = Path(file_path)
    gw_file = file.parent / f'gw_{file.stem}.py'

    gw_file.with_suffix('.yaml').unlink()
    gw_file.with_suffix('.job.yaml').unlink()
    gw_file.with_suffix('.cwl').unlink()
    image_name = f"{repository}/claimed-gw-{file_path.rsplit('.')[0].replace('_', '-')}:test"
    # TODO: Modify subprocess call to test grid wrapper
    # subprocess.run(['docker', 'run', image_name], check=True)



import json
import re
import os
import logging
from c3.parser import ContentParser, NotebookReader


class Notebook():
    def __init__(self, path):
        self.path = path
        with open(path) as json_file:
            self.notebook = json.load(json_file)

        self.name = os.path.basename(path)[:-6].replace('_', '-').lower()

        if self.notebook['cells'][1]['cell_type'] == self.notebook['cells'][2]['cell_type'] == 'markdown':
            # backwards compatibility (v0.1 description was included in second cell, merge first two markdown cells)
            logging.info('Merge first two markdown cells for description. '
                         'The file name is used as the operator name, not the first markdown cell.')
            self.description = self.notebook['cells'][1]['source'][0] + '\n' + self.notebook['cells'][2]['source'][0]
        else:
            # Using second cell because first cell was added for setup code
            self.description = self.notebook['cells'][1]['source'][0]

        self.inputs = self._get_input_vars()
        self.outputs = self._get_output_vars()

    def _get_input_vars(self):
        cp = ContentParser()
        env_names = cp.parse(self.path)['inputs']
        return_value = dict()
        notebook_code_lines = list(NotebookReader(self.path).read_next_code_line())
        for env_name, default in env_names.items():
            comment_line = str()
            for line in notebook_code_lines:
                if re.search("[\"']" + env_name + "[\"']", line):
                    if not comment_line.strip().startswith('#'):
                        # previous line was no description, reset comment_line.
                        comment_line = ''
                    if comment_line == '':
                        logging.info(f'Interface: No description for variable {env_name} provided.')
                    if re.search(r'=\s*int\(\s*os', line):
                        type = 'Integer'
                    elif re.search(r'=\s*float\(\s*os', line):
                        type = 'Float'
                    elif re.search(r'=\s*bool\(\s*os', line):
                        type = 'Boolean'
                    else:
                        type = 'String'
                    return_value[env_name] = {
                        'description': comment_line.replace('#', '').replace("\"", "\'").strip(),
                        'type': type,
                        'default': default
                    }
                    break
                comment_line = line
        return return_value

    def _get_output_vars(self):
        cp = ContentParser()
        output_names = cp.parse(self.path)['outputs']
        # TODO: Does not check for description code
        return_value = {name: {
            'description': f'Output path for {name}',
            'type': 'String',
        } for name in output_names}
        return return_value

    def get_requirements(self):
        requirements = []
        notebook_code_lines = list(NotebookReader(self.path).read_next_code_line())
        # Add dnf install
        for line in notebook_code_lines:
            if re.search(r'[\s#]*dnf\s*.[^#]*', line):
                if '-y' not in line:
                    # Adding default repo
                    line += ' -y'
                requirements.append(line.replace('#', '').strip())

        # Add pip install
        pattern = r"^[# ]*(pip[ ]*install)[ ]*(.[^#]*)"
        for line in notebook_code_lines:
            result = re.findall(pattern, line)
            if len(result) == 1:
                requirements.append((result[0][0] + ' ' + result[0][1].strip()))
        return requirements

    def get_name(self):
        return self.name

    def get_description(self):
        return self.description

    def get_inputs(self):
        return self.inputs

    def get_outputs(self):
        return self.outputs




import os
import sys
import logging
import shutil
import argparse
import subprocess
import glob
import re
import json
from pathlib import Path
from string import Template
from typing import Optional
from c3.pythonscript import Pythonscript
from c3.notebook import Notebook
from c3.rscript import Rscript
from c3.utils import convert_notebook, get_image_version
from c3.templates import (python_component_setup_code, component_setup_code_wo_logging, r_component_setup_code,
                          python_dockerfile_template, r_dockerfile_template,
                          kfp_component_template, kubernetes_job_template, cwl_component_template)

CLAIMED_VERSION = 'V0.1'


def create_dockerfile(dockerfile_template, requirements, target_code, target_dir, additional_files, working_dir, command):
    # Check for requirements file
    for i in range(len(requirements)):
        if '-r ' in requirements[i]:
            r_file_search = re.search('-r ~?\/?([A-Za-z0-9\/]*\.txt)', requirements[i])
            if len(r_file_search.groups()):
                # Get file from regex
                requirements_file = r_file_search.groups()[0]
                if requirements_file not in additional_files and os.path.isfile(requirements_file):
                    # Add missing requirements text file to additional files
                    additional_files.append(r_file_search.groups()[0])
            if '/' not in requirements[i]:
                # Add missing home directory to the command `pip install -r ~/requirements.txt`
                requirements[i] = requirements[i].replace('-r ', '-r ~/')

    requirements_docker = list(map(lambda s: 'RUN ' + s, requirements))
    requirements_docker = '\n'.join(requirements_docker)
    additional_files_docker = list(map(lambda s: f"ADD {s} {working_dir}{s}", additional_files))
    additional_files_docker = '\n'.join(additional_files_docker)

    docker_file = dockerfile_template.substitute(
        requirements_docker=requirements_docker,
        target_code=target_code,
        target_dir=target_dir,
        additional_files_docker=additional_files_docker,
        working_dir=working_dir,
        command=os.path.basename(command),
    )

    logging.info('Create Dockerfile')
    with open("Dockerfile", "w") as text_file:
        text_file.write(docker_file)
    logging.debug('Dockerfile:\n' + docker_file)


def create_kfp_component(name, description, repository, version, command, target_code, target_dir, file_path, inputs, outputs):

    inputs_list = str()
    for name, options in inputs.items():
        inputs_list += f'- {{name: {name}, type: {options["type"]}, description: "{options["description"]}"'
        if options['default'] is not None:
            if not options["default"].startswith('"'):
                options["default"] = f'"{options["default"]}"'
            inputs_list += f', default: {options["default"]}'
        inputs_list += '}\n'

    outputs_list = str()
    for name, options in outputs.items():
        outputs_list += f'- {{name: {name}, type: String, description: "{options["description"]}"}}\n'

    parameter_list = str()
    for index, key in enumerate(list(inputs.keys()) + list(outputs.keys())):
        parameter_list += f'{key}="${{{index}}}" '

    parameter_values = str()
    for input_key in inputs.keys():
        parameter_values += f"        - {{inputValue: {input_key}}}\n"
    for output_key in outputs.keys():
        parameter_values += f"        - {{outputPath: {output_key}}}\n"

    yaml = kfp_component_template.substitute(
        name=name,
        description=description,
        repository=repository,
        version=version,
        inputs=inputs_list,
        outputs=outputs_list,
        command=os.path.basename(command),
        target_dir=target_dir,
        target_code=target_code,
        parameter_list=parameter_list,
        parameter_values=parameter_values,
    )

    logging.debug('KubeFlow component yaml:\n' + yaml)
    target_yaml_path = str(Path(file_path).with_suffix('.yaml'))

    logging.info(f'Write KubeFlow component yaml to {target_yaml_path}')
    with open(target_yaml_path, "w") as text_file:
        text_file.write(yaml)


def create_kubernetes_job(name, repository, version, target_code, target_dir, command, working_dir, file_path, inputs):
    # get environment entries
    env_entries = str()
    for key in list(inputs.keys()):
        env_entries += f"        - name: {key}\n          value: value_of_{key}\n"
    env_entries = env_entries.rstrip()

    job_yaml = kubernetes_job_template.substitute(
        name=name,
        repository=repository,
        version=version,
        target_code=target_code,
        target_dir=target_dir,
        env_entries=env_entries,
        command=command,
        working_dir=working_dir,
    )

    logging.debug('Kubernetes job yaml:\n' + job_yaml)
    target_job_yaml_path = str(Path(file_path).with_suffix('.job.yaml'))

    logging.info(f'Write kubernetes job yaml to {target_job_yaml_path}')
    with open(target_job_yaml_path, "w") as text_file:
        text_file.write(job_yaml)


def create_cwl_component(name, repository, version, file_path, inputs, outputs):
    type_dict = {'String': 'string', 'Integer': 'int', 'Float': 'float', 'Boolean': 'bool'}
    # get environment entries
    i = 1
    input_envs = str()
    for input, options in inputs.items():
        i += 1
        # Convert string default value to CWL types
        default_value = options['default'] if options['type'] == 'String' and options['default'] != '"None"' \
            else options['default'].strip('"\'')
        input_envs += (f"  {input}:\n    type: {type_dict[options['type']]}\n    default: {default_value}\n    "
                       f"inputBinding:\n      position: {i}\n      prefix: --{input}\n")

    if len(outputs) == 0:
        output_envs = '[]'
    else:
        output_envs = '\n'
    for output, options in outputs.items():
        i += 1
        output_envs += (f"  {output}:\n    type: string\n    "
                        f"inputBinding:\n      position: {i}\n      prefix: --{output}\n")

    cwl = cwl_component_template.substitute(
        name=name,
        repository=repository,
        version=version,
        inputs=input_envs,
        outputs=output_envs,
    )

    logging.debug('CWL component:\n' + cwl)
    target_cwl_path = str(Path(file_path).with_suffix('.cwl'))

    logging.info(f'Write cwl component to {target_cwl_path}')
    with open(target_cwl_path, "w") as text_file:
        text_file.write(cwl)


def check_existing_files(file_path, rename_files, overwrite_files):
    if rename_files is None and overwrite_files:
        # Overwrite potential files
        return

    target_job_yaml_path = Path(file_path).with_suffix('.job.yaml')

    # Check for existing job yaml
    if target_job_yaml_path.is_file():
        if rename_files is None:
            # Ask user
            rename_files = input(f'\nFound a existing Kubernetes job file at {target_job_yaml_path}.\n'                      
                                 f'ENTER to overwrite the file, write Y to rename the file to '
                                 f'modified_{target_job_yaml_path.name}, or provide a custom name:\n')
        if rename_files.strip() == '':
            # Overwrite file
            return
        elif rename_files.lower() == 'y':
            # Default file name
            new_file_name = 'modified_' + Path(file_path).name
        else:
            # Rename to custom name
            new_file_name = rename_files

        modified_path = (target_job_yaml_path.parent / new_file_name).with_suffix('.job.yaml')
        # Check if modified path exists and potentially overwrite
        if modified_path.exists():
            if overwrite_files:
                logging.info(f'Overwriting modified path {modified_path}.')
            else:
                overwrite = input(f'Modified path {modified_path} already exists. ENTER to overwrite the file.')
                if overwrite != '':
                    logging.error(f'Abort creating operator. Please rename file manually and rerun the script.')
                    raise FileExistsError

        os.rename(str(target_job_yaml_path), str(modified_path))
        logging.info(f'Renamed Kubernetes job file to {modified_path}')
    # TODO: Should we check other files too? Currently assuming no modification for yaml and cwl.


def print_claimed_command(name, repository, version, inputs):
    claimed_command = f"claimed --component {repository}/claimed-{name}:{version}"
    for input, options in inputs.items():
        claimed_command += f" --{input} {options['default']}"
    logging.info(f'Run operators locally with claimed-cli:\n{claimed_command}')


def remove_temporary_files(file_path, target_code):
    logging.info(f'Remove local files')
    # remove temporary files
    if file_path != target_code:
        os.remove(target_code)
    if os.path.isfile('Dockerfile'):
        os.remove('Dockerfile')


def create_operator(file_path: str,
                    repository: str,
                    version: str,
                    custom_dockerfile_template: Optional[Template],
                    additional_files: str = None,
                    log_level='INFO',
                    local_mode=False,
                    no_cache=False,
                    rename_files=None,
                    overwrite_files=False,
                    skip_logging=False,
                    keep_generated_files=False,
                    ):
    logging.info('Parameters: ')
    logging.info('file_path: ' + file_path)
    logging.info('repository: ' + str(repository))
    logging.info('version: ' + str(version))
    logging.info('additional_files: ' + '; '.join(additional_files))

    if file_path.endswith('.py'):
        # use temp file for processing
        target_code = 'claimed_' + os.path.basename(file_path)
        # Copy file to current working directory
        shutil.copy(file_path, target_code)
        # Add code for logging and cli parameters to the beginning of the script
        with open(target_code, 'r') as f:
            script = f.read()
        if skip_logging:
            script = component_setup_code_wo_logging + script
        else:
            script = python_component_setup_code + script
        with open(target_code, 'w') as f:
            f.write(script)
        # getting parameter from the script
        script_data = Pythonscript(target_code)
        dockerfile_template = custom_dockerfile_template or python_dockerfile_template
        command = '/opt/app-root/bin/python'
        working_dir = '/opt/app-root/src/'

    elif file_path.endswith('.ipynb'):
        # use temp file for processing
        target_code = 'claimed_' + os.path.basename(file_path)
        # Copy file to current working directory
        shutil.copy(file_path, target_code)
        with open(target_code, 'r') as json_file:
            notebook = json.load(json_file)
        # Add code for logging and cli parameters to the beginning of the notebook
        notebook['cells'].insert(0, {
            'cell_type': 'code', 'execution_count': None, 'metadata': {}, 'outputs': [],
            'source': component_setup_code_wo_logging if skip_logging else python_component_setup_code})
        with open(target_code, 'w') as json_file:
             json.dump(notebook, json_file)
        # getting parameter from the script
        script_data = Notebook(target_code)
        dockerfile_template = custom_dockerfile_template or python_dockerfile_template
        command = '/opt/app-root/bin/ipython'
        working_dir = '/opt/app-root/src/'

    elif file_path.lower().endswith('.r'):
        # use temp file for processing
        target_code = 'claimed_' + os.path.basename(file_path)
        # Copy file to current working directory
        shutil.copy(file_path, target_code)
        # Add code for logging and cli parameters to the beginning of the script
        with open(target_code, 'r') as f:
            script = f.read()
        script = r_component_setup_code + script
        with open(target_code, 'w') as f:
            f.write(script)
        # getting parameter from the script
        script_data = Rscript(target_code)
        dockerfile_template = custom_dockerfile_template or r_dockerfile_template
        command = 'Rscript'
        working_dir = '/home/docker/'
    else:
        raise NotImplementedError('Please provide a file_path to a jupyter notebook, python script, or R script.')

    name = script_data.get_name()
    # convert description into a string with a single line
    description = ('"' + script_data.get_description().replace('\n', ' ').replace('"', '\'') +
                   ' – CLAIMED ' + CLAIMED_VERSION + '"')
    inputs = script_data.get_inputs()
    outputs = script_data.get_outputs()
    requirements = script_data.get_requirements()
    # Strip 'claimed-' from name of copied temp file
    if name.startswith('claimed-'):
        name = name[8:]
    target_dir = os.path.dirname(file_path)
    # Check that the main file is within the cwd
    if '../' in target_dir:
        raise PermissionError(f"Forbidden path outside the docker build context: {target_dir}. "
                              f"Change the current working directory to include the file.")
    elif target_dir != '':
        target_dir += '/'

    logging.info('Operator name: ' + name)
    logging.info('Description: ' + description)
    logging.info('Inputs:\n' + ('\n'.join([f'{k}: {v}' for k, v in inputs.items()])))
    logging.info('Outputs:\n' + ('\n'.join([f'{k}: {v}' for k, v in outputs.items()])))
    logging.info('Requirements: ' + '; '.join(requirements))
    logging.debug(f'Target code: {target_code}')
    logging.debug(f'Target directory: {target_dir}')

    # Load all additional files
    logging.debug('Looking for additional files:')
    additional_files_found = []
    for file_pattern in additional_files:
        if '../' in file_pattern:
            # Check that additional file are within the cwd
            raise PermissionError(f"Forbidden path outside the docker build context: {file_pattern}. "
                                  f"Change the current working directory to include all additional files.")
        # Include files based on wildcards
        files_found = glob.glob(file_pattern)
        if len(files_found) == 0:
            raise FileNotFoundError(f'No additional files for path {file_pattern}.')
        additional_files_found.extend(files_found)
        logging.debug(f'Searched for "{file_pattern}". Found {", ".join(files_found)}')
    logging.info(f'Found {len(additional_files_found)} additional files and directories\n'
                 f'{", ".join(additional_files_found)}')

    create_dockerfile(dockerfile_template, requirements, target_code, target_dir, additional_files_found, working_dir,
                      command)

    if version is None:
        # auto increase version based on registered images
        version = get_image_version(repository, name)

    if repository is None:
        if not local_mode:
            logging.warning('No repository provided. The container image is only saved locally. Add `-r <repository>` '
                            'to push the image to a container registry or run `--local_mode` to suppress this warning.')
        local_mode = True

    logging.info(f'Building container image claimed-{name}:{version}')
    try:
        # Run docker build
        subprocess.run(
            f"docker build --platform linux/amd64 -t claimed-{name}:{version} . {'--no-cache' if no_cache else ''}",
            stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True, shell=True
        )
        if repository is not None:
            # Run docker tag
            logging.debug(f'Tagging images with "latest" and "{version}"')
            subprocess.run(
                f"docker tag claimed-{name}:{version} {repository}/claimed-{name}:{version}",
                stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True, shell=True,
            )
            subprocess.run(
                f"docker tag claimed-{name}:{version} {repository}/claimed-{name}:latest",
                stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True, shell=True,
            )
    except Exception as err:
        logging.error('Docker build failed. Consider running C3 with `--log_level DEBUG` to see the docker build logs.')
        if not keep_generated_files:
            remove_temporary_files(file_path, target_code)
        raise err
    logging.info(f'Successfully built image claimed-{name}:{version}')

    if local_mode:
        logging.info(f'No repository provided, skip docker push.')
    else:
        logging.info(f'Pushing images to registry {repository}')
        try:
            # Run docker push
            subprocess.run(
                f"docker push {repository}/claimed-{name}:latest",
                stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True, shell=True,
            )
            subprocess.run(
                f"docker push {repository}/claimed-{name}:{version}",
                stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True, shell=True,
            )
            logging.info('Successfully pushed image to registry')
        except Exception as err:
            logging.error(f'Could not push images to namespace {repository}. '
                          f'Please check if docker is logged in or select a namespace with access.')
            if not keep_generated_files:
                remove_temporary_files(file_path, target_code)
            raise err

    # Check for existing files and optionally modify them before overwriting
    try:
        check_existing_files(file_path, rename_files, overwrite_files)
    except Exception as err:
        if not keep_generated_files:
            remove_temporary_files(file_path, target_code)
        raise err

    # Create application scripts
    create_kfp_component(name, description, repository, version, command, target_code, target_dir, file_path, inputs,
                         outputs)

    create_kubernetes_job(name, repository, version, target_code, target_dir, command, working_dir, file_path, inputs)

    create_cwl_component(name, repository, version, file_path, inputs, outputs)

    print_claimed_command(name, repository, version, inputs)

    # Remove temp files
    if not keep_generated_files:
        remove_temporary_files(file_path, target_code)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('FILE_PATH', type=str,
                        help='Path to python script or notebook')
    parser.add_argument('ADDITIONAL_FILES', type=str, nargs='*',
                        help='Paths to additional files to include in the container image')
    parser.add_argument('-r', '--repository', type=str, default=None,
                        help='Container registry address, e.g. docker.io/<username>')
    parser.add_argument('-v', '--version', type=str, default=None,
                        help='Container image version. Auto-increases the version number if not provided (default 0.1)')
    parser.add_argument('--rename', type=str, nargs='?', default=None, const='',
                        help='Rename existing yaml files (argument without value leads to modified_{file name})')
    parser.add_argument('--overwrite', action='store_true', help='Overwrite existing yaml files')
    parser.add_argument('-l', '--log_level', type=str, default='INFO')
    parser.add_argument('--dockerfile_template_path', type=str, default='',
                        help='Path to custom dockerfile template')
    parser.add_argument('--local_mode', action='store_true',
                        help='Continue processing after docker errors.')
    parser.add_argument('--no-cache', action='store_true', help='Not using cache for docker build.')
    parser.add_argument('--skip-logging', action='store_true',
                        help='Exclude logging code from component setup code')
    parser.add_argument('--keep-generated-files', action='store_true',
                        help='Do not delete temporary generated files.')
    args = parser.parse_args()

    # Init logging
    root = logging.getLogger()
    root.setLevel(args.log_level)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    handler.setLevel(args.log_level)
    root.addHandler(handler)

    # Update dockerfile template if specified
    if args.dockerfile_template_path != '':
        logging.info(f'Uses custom dockerfile template from {args.dockerfile_template_path}')
        with open(args.dockerfile_template_path, 'r') as f:
            custom_dockerfile_template = Template(f.read())
    else:
        custom_dockerfile_template = None

    create_operator(
        file_path=args.FILE_PATH,
        repository=args.repository,
        version=args.version,
        custom_dockerfile_template=custom_dockerfile_template,
        additional_files=args.ADDITIONAL_FILES,
        log_level=args.log_level,
        local_mode=args.local_mode,
        no_cache=args.no_cache,
        overwrite_files=args.overwrite,
        rename_files=args.rename,
        skip_logging=args.skip_logging,
        keep_generated_files=args.keep_generated_files,
    )


if __name__ == '__main__':
    main()



import os
import logging
import nbformat
import re
import subprocess
from nbconvert.exporters import PythonExporter


def convert_notebook(path):
    notebook = nbformat.read(path, as_version=4)

    # backwards compatibility (v0.1 description was included in second cell, merge first two markdown cells)
    if notebook['cells'][0]['cell_type'] == 'markdown' and notebook['cells'][1]['cell_type'] == 'markdown':
        logging.info('Merge first two markdown cells. File name is used as operator name, not first markdown cell.')
        notebook['cells'][1]['source'] = notebook['cells'][0]['source'] + '\n' + notebook['cells'][1]['source']
        notebook['cells'] = notebook['cells'][1:]

    for cell in notebook['cells']:
        if cell['cell_type'] == 'markdown':
            # convert markdown to doc string
            cell['cell_type'] = 'code'
            cell['source'] = '"""\n' + cell['source'] + '\n"""'
            cell['outputs'] = []
            cell['execution_count'] = 0
        if cell['cell_type'] == 'code' and re.search('![ ]*pip', cell['source']):
            # replace !pip with #pip
            cell['source'] = re.sub('![ ]*pip[ ]*install', '# pip install', cell['source'])

    # convert tp python script
    (code, _) = PythonExporter().from_notebook_node(notebook)

    # add import get_ipython
    code = 'from IPython import get_ipython \n' + code

    py_path = path.split('/')[-1].replace('.ipynb', '.py').replace('-', '_')

    assert not os.path.exists(py_path), f"File {py_path} already exist. Cannot convert notebook."
    with open(py_path, 'w') as py_file:
        py_file.write(code)

    return py_path


def increase_image_version(last_version):
    try:
        # increase last version value by 1
        version = last_version.split('.')
        version[-1] = str(int(version[-1]) + 1)
        version = '.'.join(version)
    except:
        # fails if a string value was used for the last tag
        version = last_version + '.1'
        logging.debug(f'Failed to increase last value, adding .1')
        pass
    return version


def pull_docker_image_tags(image):
    logging.warning("The current implementation can only query local docker images. "
                    "Please use an argument '-v <version>' to avoid duplicates.")
    # TODO: Add script for reading image tags from docker hub
    # list images
    output = subprocess.run(
        ['docker', 'image', 'ls', image],
        stdout=subprocess.PIPE
    ).stdout.decode('utf-8')
    try:
        # remove header
        image_list = output.splitlines()[1:]
        # get list of image tags
        image_tags = [line.split()[1] for line in image_list]
    except:
        image_tags = []
        logging.error(f"Could not load image tags from 'docker image ls' output: {output}")
        pass

    # filter latest and none
    image_tags = [t for t in image_tags if t not in ['latest', '<none>']]
    return image_tags


def pull_icr_image_tags(image):
    # list images from icr
    output = subprocess.run(
        ['ibmcloud', 'cr', 'images', '--restrict', image.split('icr.io/', 1)[1]],
        stdout=subprocess.PIPE
    ).stdout.decode('utf-8')

    try:
        assert 'You have no images in the namespaces' not in output
        # remove header and final status
        image_list = output.splitlines()[3:-2]
        # get list of image tags
        image_tags = [line.split()[1] for line in image_list]
    except:
        image_tags = []
        logging.warning(f"Could not load image tags from 'ibmcloud cr images' output: {output}")
        pass

    # filter latest and none
    image_tags = [t for t in image_tags if t not in ['latest', '<none>']]
    return image_tags


def get_image_version(repository, name):
    """
    Get current version of the image from the registry and increase the version by 1.
    Defaults to 0.1 if no image is found in the registry.
    """
    if repository is None:
        logging.debug('Using 0.1 as local version.')
        return '0.1'

    logging.debug(f'Get image version from registry.')
    if 'docker.io' in repository:
        logging.debug('Get image tags from docker.')
        image_tags = pull_docker_image_tags(f'{repository}/claimed-{name}')
    elif 'icr.io' in repository:
        logging.debug('Get image tags from ibmcloud container registry.')
        image_tags = pull_icr_image_tags(f'{repository}/claimed-{name}')
    else:
        logging.warning('Unrecognised container registry, using docker to query image tags.')
        image_tags = pull_docker_image_tags(f'{repository}/claimed-{name}')
    logging.debug(f'Image tags: {image_tags}')

    def check_only_numbers(test_str):
        return set(test_str) <= set('.0123456789')

    if len(image_tags) == 0:
        # default version
        version = '0.1'
        logging.info(f'Using default version {version}. No prior image tag found for {repository}/claimed-{name}.')

    elif not check_only_numbers(image_tags[0]):
        # increase last version
        version = increase_image_version(image_tags[0])
        logging.info(f'Using version {version} based on last version {image_tags[0]}.')

    else:
        # find the highest numerical version
        image_tags = list(filter(check_only_numbers, image_tags))
        image_tags.sort(key=lambda s: list(map(int, s.split('.'))))
        version = increase_image_version(image_tags[-1])
        logging.info(f'Using version {version} based on highest previous version {image_tags[-1]}.')

    return version




import logging
import os
import argparse
import sys
from string import Template
from c3.pythonscript import Pythonscript
from c3.utils import convert_notebook
from c3.create_operator import create_operator
from c3.templates import grid_wrapper_template, cos_grid_wrapper_template, component_setup_code_wo_logging


def wrap_component(component_path,
                   component_description,
                   component_dependencies,
                   component_interface,
                   component_inputs,
                   component_process,
                   cos,
                   ):
    # get component name from path
    component_name = os.path.splitext(os.path.basename(component_path))[0]

    gw_template = cos_grid_wrapper_template if cos else grid_wrapper_template
    grid_wrapper_code = gw_template.substitute(
        component_name=component_name,
        component_description=component_description,
        component_dependencies=component_dependencies,
        component_inputs=component_inputs,
        component_interface=component_interface,
        component_process=component_process,
    )

    # Write edited code to file
    grid_wrapper_file = f'cgw_{component_name}.py' if cos else f'gw_{component_name}.py'
    grid_wrapper_file_path = os.path.join(os.path.dirname(component_path), grid_wrapper_file)
    # remove 'component_' from gw path
    grid_wrapper_file_path = grid_wrapper_file_path.replace('component_', '')
    with open(grid_wrapper_file_path, 'w') as f:
        f.write(grid_wrapper_code)

    logging.info(f'Saved wrapped component to {grid_wrapper_file_path}')

    return grid_wrapper_file_path


def get_component_elements(file_path):
    # get required elements from component code
    py = Pythonscript(file_path)
    # convert description into a string with a single line
    description = (py.get_description().replace('\n', ' ').replace('"', '\''))
    inputs = py.get_inputs()
    outputs = py.get_outputs()
    dependencies = py.get_requirements()

    # combine dependencies list
    dependencies = '\n# '.join(dependencies)

    # generate interface code from inputs
    interface = ''
    type_to_func = {'String': '', 'Boolean': 'bool', 'Integer': 'int', 'Float': 'float'}
    for variable, d in inputs.items():
        interface += f"# {d['description']}\n"
        if d['type'] == 'String' and d['default'] is not None and d['default'][0] not in '\'\"':
            # Add quotation marks
            d['default'] = "'" + d['default'] + "'"
        interface += f"component_{variable} = {type_to_func[d['type']]}(os.getenv('{variable}', {d['default']}))\n"

    # TODO: Implement output interface
    if len(outputs) > 0:
        logging.warning('Found output paths in the component code which is currently not supported.')

    # generate kwargs for the subprocesses
    process_inputs = ', '.join([f'{i}=component_{i}' for i in inputs.keys()])
    # use log level from grid wrapper
    process_inputs = process_inputs.replace('component_log_level', 'log_level')

    return description, interface, process_inputs, dependencies


# Adding code
def edit_component_code(file_path, component_process):
    file_name = os.path.basename(file_path)
    if file_path.endswith('.ipynb'):
        logging.info('Convert notebook to python script')
        target_file = convert_notebook(file_path)
        file_path = target_file
        file_name = os.path.basename(file_path)
    else:
        # write edited code to different file
        target_file = os.path.join(os.path.dirname(file_path), 'component_' + file_name.replace('-', '_'))

    target_file_name = os.path.basename(target_file)

    with open(file_path, 'r') as f:
        script = f.read()
    assert component_process in script, (f'Did not find the grid process {component_process} in the script. '
                                         f'Please provide the grid process in the arguments `-p <grid_process>`.')
    # Add code for logging and cli parameters to the beginning of the script
    script = component_setup_code_wo_logging + script
    # replace old filename with new file name
    script = script.replace(file_name, target_file_name)
    with open(target_file, 'w') as f:
        f.write(script)

    if '__main__' not in script:
        logging.warning('No __main__ found in component code. Grid wrapper will import functions from component, '
                        'which can lead to unexpected behaviour without using __main__.')

    logging.info('Saved component python script in ' + target_file)

    return target_file


def apply_grid_wrapper(file_path, component_process, cos):
    assert file_path.endswith('.py') or file_path.endswith('.ipynb'), \
        "Please provide a component file path to a python script or notebook."

    file_path = edit_component_code(file_path, component_process)

    description, interface, inputs, dependencies = get_component_elements(file_path)

    component_elements = dict(
        component_path=file_path,
        component_description=description,
        component_dependencies=dependencies,
        component_interface=interface,
        component_inputs=inputs,
        component_process=component_process
    )

    logging.debug('Wrap component with parameters:')
    for component, value in component_elements.items():
        logging.debug(component + ':\n' + str(value) + '\n')

    logging.info('Wrap component')
    grid_wrapper_file_path = wrap_component(cos=cos, **component_elements)
    return grid_wrapper_file_path, file_path


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('FILE_PATH', type=str,
                        help='Path to python script or notebook')
    parser.add_argument('ADDITIONAL_FILES', type=str, nargs='*',
                        help='List of paths to additional files to include in the container image')
    parser.add_argument('-p', '--component_process', type=str, default='grid_process',
                        help='Name of the component sub process that is executed for each batch.')
    parser.add_argument('--cos', action=argparse.BooleanOptionalAction, default=False,
                        help='Creates a grid wrapper for processing COS files')
    parser.add_argument('-r', '--repository', type=str, default=None,
                        help='Container registry address, e.g. docker.io/<username>')
    parser.add_argument('-v', '--version', type=str, default=None,
                        help='Container image version. Auto-increases the version number if not provided (default 0.1)')
    parser.add_argument('--rename', type=str, nargs='?', default=None, const='',
                        help='Rename existing yaml files (argument without value leads to modified_{file name})')
    parser.add_argument('--overwrite', action='store_true', help='Overwrite existing yaml files')
    parser.add_argument('-l', '--log_level', type=str, default='INFO')
    parser.add_argument('--dockerfile_template_path', type=str, default='',
                        help='Path to custom dockerfile template')
    parser.add_argument('--local_mode', action='store_true',
                        help='Continue processing after docker errors.')
    parser.add_argument('--no-cache', action='store_true', help='Not using cache for docker build.')
    parser.add_argument('--skip-logging', action='store_true',
                        help='Exclude logging code from component setup code')
    parser.add_argument('--keep-generated-files', action='store_true',
                        help='Do not delete temporary generated files.')

    args = parser.parse_args()

    # Init logging
    root = logging.getLogger()
    root.setLevel(args.log_level)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    handler.setLevel(args.log_level)
    root.addHandler(handler)

    grid_wrapper_file_path = component_path = ''
    try:
        grid_wrapper_file_path, component_path = apply_grid_wrapper(
            file_path=args.FILE_PATH,
            component_process=args.component_process,
            cos=args.cos,
        )

        logging.info('Generate CLAIMED operator for grid wrapper')

        # Add component path and init file path to additional_files
        args.ADDITIONAL_FILES.append(component_path)

        # Update dockerfile template if specified
        if args.dockerfile_template_path != '':
            logging.info(f'Uses custom dockerfile template from {args.dockerfile_template_path}')
            with open(args.dockerfile_template_path, 'r') as f:
                custom_dockerfile_template = Template(f.read())
        else:
            custom_dockerfile_template = None

        create_operator(
            file_path=grid_wrapper_file_path,
            repository=args.repository,
            version=args.version,
            custom_dockerfile_template=custom_dockerfile_template,
            additional_files=args.ADDITIONAL_FILES,
            log_level=args.log_level,
            local_mode=args.local_mode,
            no_cache=args.no_cache,
            overwrite_files=args.overwrite,
            rename_files=args.rename,
            skip_logging=args.skip_logging,
            keep_generated_files=args.keep_generated_files,
        )
    except Exception as err:
        logging.error('Error while generating CLAIMED grid wrapper. '
                      'Consider using `--log_level DEBUG` and `--keep-generated-files` for debugging.')
        raise err
    finally:
        if not args.keep_generated_files:
            logging.info('Remove local component file and grid wrapper code.')
            if os.path.isfile(grid_wrapper_file_path):
                os.remove(grid_wrapper_file_path)
            if os.path.isfile(component_path):
                os.remove(component_path)


if __name__ == '__main__':
    main()




import logging
import os
import re
from c3.parser import ContentParser


class Rscript:
    def __init__(self, path):

        self.path = path
        with open(path, 'r') as f:
            self.script = f.read()

        self.name = os.path.basename(path)[:-2].replace('_', '-').lower()
        # TODO: Currently does not support a description
        self.description = self.name
        self.inputs = self._get_input_vars()
        self.outputs = self._get_output_vars()

    def _get_input_vars(self):
        cp = ContentParser()
        env_names = cp.parse(self.path)['inputs']
        return_value = dict()
        for env_name, default in env_names.items():
            comment_line = str()
            for line in self.script.split('\n'):
                if re.search("[\"']" + env_name + "[\"']", line):
                    # Check the description for current variable
                    if not comment_line.strip().startswith('#'):
                        # previous line was no description, reset comment_line.
                        comment_line = ''
                    if comment_line == '':
                        logging.info(f'Interface: No description for variable {env_name} provided.')
                    if re.search(r'=\s*as.numeric\(\s*os', line):
                        type = 'Float'  # double in R
                    elif re.search(r'=\s*bool\(\s*os', line):
                        type = 'Boolean'  # logical in R
                    else:
                        type = 'String'  # character in R

                    return_value[env_name] = {
                        'description': comment_line.replace('#', '').replace("\"", "\'").strip(),
                        'type': type,
                        'default': default
                    }
                    break
                comment_line = line
        return return_value

    def _get_output_vars(self):
        cp = ContentParser()
        output_names = cp.parse(self.path)['outputs']
        # TODO: Does not check for description
        return_value = {name: {'description': 'output path'} for name in output_names}
        return return_value

    def get_requirements(self):
        requirements = []
        # Add apt install commands
        for line in self.script.split('\n'):
            if re.search(r'[\s#]*apt\s*[A-Za-z0-9_-]*', line):
                if '-y' not in line:
                    # Adding default repo
                    line += ' -y'
                requirements.append(line.replace('#', '').strip())

        # Add Rscript install.packages commands
        for line in self.script.split('\n'):
            if re.search(r'[\s#]*install\.packages\(.*\)', line):
                if 'http://' not in line:
                    # Adding default repo
                    line = line.rstrip(') ') + ", repos='http://cran.us.r-project.org')"
                command = f"Rscript -e \"{line.replace('#', '').strip()}\""
                requirements.append(command)
        return requirements

    def get_name(self):
        return self.name

    def get_description(self):
        return self.description

    def get_inputs(self):
        return self.inputs

    def get_outputs(self):
        return self.outputs



"""
${component_name} got wrapped by cos_grid_wrapper, which wraps any CLAIMED component and implements the generic grid computing pattern for cos files https://romeokienzler.medium.com/the-generic-grid-computing-pattern-transforms-any-sequential-workflow-step-into-a-transient-grid-c7f3ca7459c8

CLAIMED component description: ${component_description}
"""

# pip install s3fs
# component dependencies
# ${component_dependencies}

import os
import json
import random
import logging
import shutil
import time
import glob
import s3fs
from datetime import datetime
from pathlib import Path

# import component code
from ${component_name} import *


# File containing batches. Provided as a comma-separated list of strings or keys in a json dict. All batch file names must contain the batch name.
gw_batch_file = os.environ.get('gw_batch_file', None)
# file path pattern like your/path/**/*.tif. Multiple patterns can be separated with commas. It is ignored if gw_batch_file is provided.
gw_file_path_pattern = os.environ.get('gw_file_path_pattern', None)
# pattern for grouping file paths into batches like ".split('.')[-2]". It is ignored if gw_batch_file is provided.
gw_group_by = os.environ.get('gw_group_by', None)

# comma-separated list of additional cos files to copy
gw_additional_source_files = os.environ.get('gw_additional_source_files', '')
# download source cos files to local input path
gw_local_input_path = os.environ.get('gw_local_input_path', 'input')
# upload local target files to target cos path
gw_local_target_path = os.environ.get('gw_local_target_path', 'target')

# cos source_access_key_id
gw_source_access_key_id = os.environ.get('gw_source_access_key_id')
# cos source_secret_access_key
gw_source_secret_access_key = os.environ.get('gw_source_secret_access_key')
# cos source_endpoint
gw_source_endpoint = os.environ.get('gw_source_endpoint')
# cos source_bucket
gw_source_bucket = os.environ.get('gw_source_bucket')

# cos target_access_key_id (uses source s3 if not provided)
gw_target_access_key_id = os.environ.get('gw_target_access_key_id', None)
# cos target_secret_access_key (uses source s3 if not provided)
gw_target_secret_access_key = os.environ.get('gw_target_secret_access_key', None)
# cos target_endpoint (uses source s3 if not provided)
gw_target_endpoint = os.environ.get('gw_target_endpoint', None)
# cos target_bucket (uses source s3 if not provided)
gw_target_bucket = os.environ.get('gw_target_bucket', None)
# cos target_path
gw_target_path = os.environ.get('gw_target_path')

# cos coordinator_access_key_id (uses source s3 if not provided)
gw_coordinator_access_key_id = os.environ.get('gw_coordinator_access_key_id', None)
# cos coordinator_secret_access_key (uses source s3 if not provided)
gw_coordinator_secret_access_key = os.environ.get('gw_coordinator_secret_access_key', None)
# cos coordinator_endpoint (uses source s3 if not provided)
gw_coordinator_endpoint = os.environ.get('gw_coordinator_endpoint', None)
# cos coordinator_bucket (uses source s3 if not provided)
gw_coordinator_bucket = os.environ.get('gw_coordinator_bucket', None)
# cos path to grid wrapper coordinator directory
gw_coordinator_path = os.environ.get('gw_coordinator_path')
# lock file suffix
gw_lock_file_suffix = os.environ.get('gw_lock_file_suffix', '.lock')
# processed file suffix
gw_processed_file_suffix = os.environ.get('gw_lock_file_suffix', '.processed')
# error file suffix
gw_error_file_suffix = os.environ.get('gw_error_file_suffix', '.err')
# timeout in seconds to remove lock file from struggling job (default 3 hours)
gw_lock_timeout = int(os.environ.get('gw_lock_timeout', 10800))
# ignore error files and rerun batches with errors
gw_ignore_error_files = bool(os.environ.get('gw_ignore_error_files', False))


# component interface
${component_interface}

# init s3
s3source = s3fs.S3FileSystem(
    anon=False,
    key=gw_source_access_key_id,
    secret=gw_source_secret_access_key,
    client_kwargs={'endpoint_url': gw_source_endpoint})

if gw_target_endpoint is not None:
    s3target = s3fs.S3FileSystem(
        anon=False,
        key=gw_target_access_key_id,
        secret=gw_target_secret_access_key,
        client_kwargs={'endpoint_url': gw_target_endpoint})
else:
    logging.debug('Using source bucket as target bucket.')
    gw_target_bucket = gw_source_bucket
    s3target = s3source

if gw_coordinator_bucket is not None:
    s3coordinator = s3fs.S3FileSystem(
        anon=False,
        key=gw_coordinator_access_key_id,
        secret=gw_coordinator_secret_access_key,
        client_kwargs={'endpoint_url': gw_coordinator_endpoint})
else:
    logging.debug('Using source bucket as coordinator bucket.')
    gw_coordinator_bucket = gw_source_bucket
    s3coordinator = s3source

def load_batches_from_file(batch_file):
    if batch_file.endswith('.json'):
        # load batches from keys of a json file
        logging.info(f'Loading batches from json file: {batch_file}')
        with s3source.open(Path(gw_source_bucket) / batch_file, 'r') as f:
            batch_dict = json.load(f)
        batches = batch_dict.keys()

    else:
        # Load batches from comma-separated txt file
        logging.info(f'Loading comma-separated batch strings from file: {batch_file}')
        with s3source.open(Path(gw_source_bucket) / batch_file, 'r') as f:
            batch_string = f.read()
        batches = [b.strip() for b in batch_string.split(',')]

    logging.info(f'Loaded {len(batches)} batches')
    logging.debug(f'List of batches: {batches}')
    assert len(batches) > 0, f"batch_file {batch_file} has no batches."
    return batches


def get_files_from_pattern(file_path_patterns):
    logging.info(f'Start identifying files')
    all_files = []

    # Iterate over comma-separated paths
    for file_path_pattern in file_path_patterns.split(','):
        logging.info(f'Get file paths from pattern: {file_path_pattern}')
        files = s3source.glob(str(Path(gw_source_bucket) / file_path_pattern.strip()))
        assert len(files) > 0, f"Found no files with file_path_pattern {file_path_pattern}."
        all_files.extend(files)
    logging.info(f'Found {len(all_files)} cos files')
    return all_files

def identify_batches_from_pattern(file_path_patterns, group_by):
    logging.info(f'Start identifying files and batches')
    batches = set()
    all_files = get_files_from_pattern(file_path_patterns)

    # get batches by applying the group by function to all file paths
    for path_string in all_files:
        part = eval('str(path_string)' + group_by, {"group_by": group_by, "path_string": path_string})
        assert part != '', f'Could not extract batch with path_string {path_string} and group_by {group_by}'
        batches.add(part)

    logging.info(f'Identified {len(batches)} batches')
    logging.debug(f'List of batches: {batches}')

    return batches, all_files


def perform_process(process, batch, cos_files):
    logging.debug(f'Check coordinator files for batch {batch}.')
    # init coordinator files
    coordinator_dir = Path(gw_coordinator_bucket) / gw_coordinator_path
    lock_file = str(coordinator_dir / (batch + gw_lock_file_suffix))
    processed_file = str(coordinator_dir / (batch + gw_processed_file_suffix))
    error_file = str(coordinator_dir / (batch + gw_error_file_suffix))

    if s3coordinator.exists(lock_file):
        # remove strugglers
        last_modified = s3coordinator.info(lock_file)['LastModified']
        if (datetime.now(last_modified.tzinfo) - last_modified).total_seconds() > gw_lock_timeout:
            logging.info(f'Lock file {lock_file} is expired.')
            s3coordinator.rm(lock_file)
        else:
            logging.debug(f'Batch {batch} is locked.')
            return

    if s3coordinator.exists(processed_file):
        logging.debug(f'Batch {batch} is processed.')
        return

    if s3coordinator.exists(error_file):
        if gw_ignore_error_files:
            logging.info(f'Ignoring previous error in batch {batch} and rerun.')
        else:
            logging.debug(f'Batch {batch} has error.')
            return

    logging.debug(f'Locking batch {batch}.')
    s3coordinator.touch(lock_file)
    logging.info(f'Processing batch {batch}.')

    # Create input and target directories
    input_path = Path(gw_local_input_path)
    target_path = Path(gw_local_target_path)
    assert not input_path.exists(), (f'gw_local_input_path ({gw_local_input_path}) already exists. '
                                     f'Please provide a new input path.')
    assert not target_path.exists(), (f'gw_local_target_path ({gw_local_target_path}) already exists. '
                                     f'Please provide a new target path.')
    input_path.mkdir(parents=True)
    target_path.mkdir(parents=True)

    # Download cos files to local input folder
    batch_fileset = list(filter(lambda file: batch in file, cos_files))
    if gw_additional_source_files != '':
        additional_source_files = [f.strip() for f in gw_additional_source_files.split(',')]
        batch_fileset.extend(additional_source_files)
    logging.info(f'Downloading {len(batch_fileset)} files from COS')
    for cos_file in batch_fileset:
        local_file = str(input_path / cos_file.split('/', 1)[-1])
        logging.debug(f'Downloading {cos_file} to {local_file}')
        s3source.get(cos_file, local_file)

    # processing files with custom process
    try:
        target_files = process(batch, ${component_inputs})
    except Exception as err:
        logging.error(f'{type(err).__name__} in batch {batch}: {err}')
        # Write error to file
        with s3coordinator.open(error_file, 'w') as f:
            f.write(f"{type(err).__name__} in batch {batch}: {err}")
        s3coordinator.rm(lock_file)
        logging.error(f'Continue processing.')
        return

    # optional verify target files
    if target_files is not None:
        if isinstance(target_files, str):
            target_files = [target_files]
        for target_file in target_files:
            if not os.path.exists(target_file):
                logging.error(f'Target file {target_file} does not exist for batch {batch}.')
        if any([not str(t).startswith(gw_local_target_path) for t in target_files]):
            logging.warning('Some target files are not in target path. Only files in target path are uploaded.')
    else:
        logging.info(f'Cannot verify batch {batch} (target files not provided). Using files in target_path.')

    # upload files in target path
    local_target_files = list(target_path.glob('*'))
    logging.info(f'Uploading {len(local_target_files)} target files to COS.')
    for local_file in local_target_files:
        cos_file = Path(gw_target_bucket) / gw_target_path / local_file.relative_to(target_path)
        logging.debug(f'Uploading {local_file} to {cos_file}')
        s3target.put(str(local_file), str(cos_file))

    logging.info(f'Remove local input and target files.')
    shutil.rmtree(input_path)
    shutil.rmtree(target_path)

    logging.info(f'Finished Batch {batch}.')
    s3coordinator.touch(processed_file)
    # Remove lock file
    if s3coordinator.exists(lock_file):
        s3coordinator.rm(lock_file)
    else:
        logging.warning(f'Lock file {lock_file} was removed by another process. '
                        f'Consider increasing gw_lock_timeout (currently {gw_lock_timeout}s) to repeated processing.')


def process_wrapper(sub_process):
    delay = random.randint(1, 60)
    logging.info(f'Staggering start, waiting for {delay} seconds')
    time.sleep(delay)

    # Init coordinator dir
    coordinator_dir = Path(gw_coordinator_bucket) / gw_coordinator_path
    s3coordinator.makedirs(coordinator_dir, exist_ok=True)

    # get batches
    if gw_batch_file is not None and os.path.isfile(gw_batch_file):
        batches = load_batches_from_file(gw_batch_file)
        cos_files = get_files_from_pattern(gw_file_path_pattern)
    elif gw_file_path_pattern is not None and gw_group_by is not None:
        batches, cos_files = identify_batches_from_pattern(gw_file_path_pattern, gw_group_by)
    else:
        raise ValueError("Cannot identify batches. "
                         "Provide valid gw_batch_file or gw_file_path_pattern and gw_group_by.")

    # Iterate over all batches
    for batch in batches:
        perform_process(sub_process, batch, cos_files)

    # Check and log status of batches
    processed_status = [s3coordinator.exists(coordinator_dir / (batch + gw_processed_file_suffix)) for batch in batches]
    lock_status = [s3coordinator.exists(coordinator_dir / (batch + gw_lock_file_suffix)) for batch in batches]
    error_status = [s3coordinator.exists(coordinator_dir / (batch + gw_error_file_suffix)) for batch in batches]

    logging.info(f'Finished current process. Status batches: '
                 f'{sum(processed_status)} processed / {sum(lock_status)} locked / {sum(error_status)} errors / {len(processed_status)} total')

    if sum(error_status):
        logging.error(f'Found errors! Resolve errors and rerun operator with gw_ignore_error_files=True.')
        # print all error messages
        for error_file in s3coordinator.glob(str(coordinator_dir / ('**/*' + gw_error_file_suffix))):
            with s3coordinator.open(error_file, 'r') as f:
                logging.error(f.read())


if __name__ == '__main__':
    process_wrapper(${component_process})



# default code for each  operator
import os
import sys
import re
import logging

# init logger
root = logging.getLogger()
root.setLevel('INFO')
handler = logging.StreamHandler(sys.stdout)
handler.setLevel('INFO')
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)
logging.basicConfig(level=logging.CRITICAL)

# get parameters from args
parameters = list(filter(
                lambda s: s.find('=') > -1 and bool(re.match(r'[A-Za-z0-9_]*=[.\/A-Za-z0-9]*', s)),
                sys.argv
            ))

# set parameters to env variables
for parameter in parameters:
    variable = parameter.split('=')[0]
    value = parameter.split('=', 1)[-1]
    logging.info(f'Parameter: {variable} = "{value}"')
    os.environ[variable] = value

# update log level
log_level = os.environ.get('log_level', 'INFO')
if log_level !='INFO':
    logging.info(f'Updating log level to {log_level}')
    root.setLevel(log_level)
    handler.setLevel(log_level)




import os
from string import Template
from pathlib import Path

# template file names
PYTHON_COMPONENT_SETUP_CODE = 'component_setup_code.py'
R_COMPONENT_SETUP_CODE = 'component_setup_code.R'
PYTHON_COMPONENT_SETUP_CODE_WO_LOGGING = 'component_setup_code_wo_logging.py'
PYTHON_DOCKERFILE_FILE = 'python_dockerfile_template'
R_DOCKERFILE_FILE = 'R_dockerfile_template'
KFP_COMPONENT_FILE = 'kfp_component_template.yaml'
KUBERNETES_JOB_FILE = 'kubernetes_job_template.job.yaml'
CWL_COMPONENT_FILE = 'cwl_component_template.cwl'
GRID_WRAPPER_FILE = 'grid_wrapper_template.py'
COS_GRID_WRAPPER_FILE = 'cos_grid_wrapper_template.py'

# load templates
template_path = Path(os.path.dirname(__file__))

with open(template_path / PYTHON_COMPONENT_SETUP_CODE, 'r') as f:
    python_component_setup_code = f.read()

with open(template_path / R_COMPONENT_SETUP_CODE, 'r') as f:
    r_component_setup_code = f.read()

with open(template_path / PYTHON_COMPONENT_SETUP_CODE_WO_LOGGING, 'r') as f:
    component_setup_code_wo_logging = f.read()

with open(template_path / PYTHON_DOCKERFILE_FILE, 'r') as f:
    python_dockerfile_template = Template(f.read())

with open(template_path / R_DOCKERFILE_FILE, 'r') as f:
    r_dockerfile_template = Template(f.read())

with open(template_path / KFP_COMPONENT_FILE, 'r') as f:
    kfp_component_template = Template(f.read())

with open(template_path / KUBERNETES_JOB_FILE, 'r') as f:
    kubernetes_job_template = Template(f.read())

with open(template_path / CWL_COMPONENT_FILE, 'r') as f:
    cwl_component_template = Template(f.read())

with open(template_path / GRID_WRAPPER_FILE, 'r') as f:
    grid_wrapper_template = Template(f.read())

with open(template_path / COS_GRID_WRAPPER_FILE, 'r') as f:
    cos_grid_wrapper_template = Template(f.read())
